"""Class module to interface with Beersmith.
"""
import os
from collections import OrderedDict

from aracnid_logger import Logger
from bs4 import BeautifulSoup
import hjson
import xmltodict

# from beersmith_direct.recipes import Recipes

# initialize logging
logger = Logger(__name__).get_logger()


class BeersmithInterface:
    """Interface to Beersmith.

    Attributes:
        bsm: 
    """

    def __init__(self) -> None:
        """Initializes the Beersmith interface.

        Args:
            None.
        """
        self.default_filename = os.environ.get('BEERSMITH_DEFAULT_FILENAME')
        self.default_path = os.environ.get('BEERSMITH_DEFAULT_PATH')

    def read_bsmx(self, filename=None, path=None):
        """Reads a .bsmx file and returns a recipe folder of recipes.

        Args:
            filename: The supplied filename.
            path: The supplied directory.

        Returns:
            A list of recipes.
        """
        self.filename = filename if filename else self.default_filename
        self.path = path if path else self.default_path

        # read the file
        filepath = os.path.join(self.path, self.filename)
        if os.path.exists(filepath):
            with(open(filepath, 'r')) as bsmx_file:
                xml_string = bsmx_file.read()

                # replace some tag names from bsmx
                xml_string = xml_string.replace('_MOD_', 'last_modified')
                xml_string = xml_string.replace('_TExpanded', 'texpanded')
                xml_string = xml_string.replace('_XName', 'xname')
                xml_string = xml_string.replace('F_R_NAME', 'name')		# why do this here

                # replace archive tag names
                xml_string = xml_string.replace('F_AR_ACTION', 'action')
                xml_string = xml_string.replace('F_AR_NAME', 'name')		# why do this here
                xml_string = xml_string.replace('F_AR_DIRECTORY', 'directory')
                xml_string = xml_string.replace('F_AR_FILE', 'file')

                # process text
                if xml_string:
                    xml_beautiful = BeautifulSoup(xml_string, 'html.parser')

                    if xml_string.startswith('<Selections>') or xml_string.startswith('<Recipe>'):
                        parsed_obj = xmltodict.parse(xml_beautiful.prettify())
                        roottag, dict_items = parsed_obj.popitem()

                        if dict_items:
                            recipe_list = self.process_recipes(dict_items)

                            return recipe_list

                    elif xml_string.startswith('<Archive>'):
                        parsed_obj = xmltodict.parse(f'<root>{xml_beautiful}</root>')
                        roottag, dict_items = parsed_obj.popitem()

                        if dict_items:
                            archive_list = self.process_archive(dict_items)

                            return archive_list

        return []

    def process_recipes(self, dict_items):
        # To handle the recursive nature of embedded folders

        # set the folder name
        folder_name = dict_items['name']

        # bounce out if folder is empty
        if dict_items['data'] is None:
            return 0

        # get the data key, could be 'table' or 'recipe'
        data_key = list(dict_items['data'].keys())[0]

        # process folder data into items list
        item_dict = self.flatten_data(dict_items, data_key, 'items')

        recipe_list = []
        item_list = item_dict['items']
        for item in item_list:
            if 'xname' in item and item['xname'] == 'Folder':
                folder_recipe_list = self.process_folder(item)
                recipe_list.extend(folder_recipe_list)
            else:
                recipe = self.process_recipe(item, folder_name=folder_name)
                recipe_list.append(recipe)

        return recipe_list

    def process_archive(self, dict_items, **kwargs):
        archive_list = dict_items['archive']

        # process as list for one action
        if type(archive_list) is OrderedDict:
            archive_list = [archive_list]

        return archive_list

    def process_folder(self, props):
        # initialize return variable
        recipe_list = []
        # props_new = props
        # props_new['_type'] = 'folder'

        folder_name = props['name']

        # flatten sub data structures
        if 'data' in props:
            recipes_dict = self.flatten_data(props, 'recipe', 'recipes')
            # process each recipe
            recipe_num = 0
            while recipe_num < len(recipes_dict['recipes']):
                recipe_props = recipes_dict['recipes'][recipe_num]
                recipe_props = self.process_recipe(recipe_props, folder_name=folder_name)
                recipe_list.append(recipe_props)
                # recipes_dict['recipes'][recipe_num] = recipe_props
                recipe_num += 1

            # # add recipe props back
            # props_new.update(recipes_dict)

        # return folder
        return recipe_list

    def process_recipe(self, props, folder_name=None):
        # initialize return variable
        props_new = {}
        props_new['_type'] = 'recipe'
        
        # add recipe id
        props['_id'] = props['name']
        #print(props['name'])

        # add correct folder name
        # print('folder_name: {}'.format(folder_name))
        props['folder_name'] = '/{}/'.format(folder_name)

        # process sub data structures, flatten and remove key prefixes
        ingredients_props, ingredients_list = self.process_ingredients(props['ingredients'].pop('data'))
        props['ingredients_by_type'] = ingredients_props
        props['ingredients'] = ingredients_list

        props['mash'] = {}
        self.strip_key_prefixes('f_mh_', props.pop('f_r_mash'), props['mash'])
        mashsteps_dict = self.flatten_data(props['mash']['steps'], 'mashstep', 'mashsteps')
        self.strip_key_prefixes_list('f_ms_', mashsteps_dict['mashsteps'])
        props['mash'].update(mashsteps_dict)

        props['equipment'] = {}
        self.strip_key_prefixes('f_e_', props.pop('f_r_equipment'), props['equipment'])

        props['style'] = {}
        self.strip_key_prefixes('f_s_', props.pop('f_r_style'), props['style'])

        props['carb'] = {}
        self.strip_key_prefixes('f_c_', props.pop('f_r_carb'), props['carb'])

        props['base_grain'] = {}
        self.strip_key_prefixes('f_g_', props.pop('f_r_base_grain'), props['base_grain'])

        props['ferment'] = {}
        self.strip_key_prefixes('f_a_', props.pop('f_r_age'), props['ferment'])
        readings_dict = self.flatten_data(props['agedata'], 'agedata', 'readings')
        self.strip_key_prefixes_list('f_ad_', readings_dict['readings'])
        props['ferment'].update(readings_dict)
        props['ferment'].update(props.pop('agedata'))

        # remove base key prefixes
        self.strip_key_prefixes('f_r_', props, props_new)

        # correct data types
        props_new = self.correct_type(props_new)

        # process notes
        notes = props_new['notes']
        if notes:
            notes_props = self.process_notes(notes)
            if type(notes_props) in (dict, OrderedDict):
                props_new.update(notes_props)

        return props_new

    def process_items(self, props, tag_old, tag_new=None):
        # initialize return variable
        props_new = {}
        items_new = []

        if not tag_new:
            tag_new = tag_old

        if props:
            # retrieve items
            items_old = props[tag_old]

            if type(items_old) == list:
                items_new.extend(items_old)
            elif type(items_old) in (dict, OrderedDict):
                items_new.append(items_old)

        # create the new properties
        props_new[tag_new] = items_new

        # return
        return props_new

    def process_ingredients(self, ingredients_props):
        # initialize return variable
        ingredients_by_type = ingredients_props
        ingredients_list = []

        # process ingredient subitems into lists
        for ingredient_type in ingredients_props:
            ingredient_dict = self.process_items(ingredients_props, ingredient_type)
            ingredient_list = ingredient_dict[ingredient_type]
            ingredient_num = 0
            while ingredient_num < len(ingredient_list):
                ingredient_props = ingredient_list[ingredient_num]
                ingredient_props = self.process_ingredient(ingredient_props, ingredient_type)
                ingredient_list[ingredient_num] = ingredient_props
                ingredients_list.append(ingredient_props)
                ingredient_num += 1

            ingredients_by_type[ingredient_type] = ingredient_list

        # sort the ingredients list
        ingredients_list = sorted(ingredients_list, key=lambda ingredient:ingredient['order'])

        return ingredients_by_type, ingredients_list

    def process_ingredient(self, ingredient_props, ingredient_type):
        # initialize new ingredient dict
        props_new = {}
        
        # determine the prefix, units
        if ingredient_type == 'grain':
            prefix = 'f_g_'
            units = 'oz'
        elif ingredient_type == 'hops':
            prefix = 'f_h_'
            units = 'oz'
        elif ingredient_type == 'yeast':
            prefix = 'f_y_'
            units = 'g'
        elif ingredient_type == 'misc':
            prefix = 'f_m_'
            units = ''
        elif ingredient_type == 'water':
            prefix = 'f_w_'
            units = 'oz'
        else:
            # print('[WARNING] Unhandled Ingredient Type: {}'.format(ingredient_type))
            prefix = ''
            units = ''

        # add 'units' tag
        props_new['units'] = units
        
        # rename f_order key
        if 'f_order' in ingredient_props:
            props_new['order'] = ingredient_props['f_order']

        # remove tag prefixes
        self.strip_key_prefixes(prefix, ingredient_props, props_new)

        # ingredient specific processing
        if ingredient_type == 'grain':
            subtype_code = self.correct_type(props_new['type'])
            subtype_list = ['grain', 'extract', 'sugar', 'adjunct', 'dry extract']
            props_new['subtype'] = subtype_list[subtype_code]

        elif ingredient_type == 'yeast':
            form_code = self.correct_type(props_new['form'])
            form_list = ['liquid', 'dry', 'slant', 'culture']
            props_new['form_text'] = form_list[form_code]
            props_new['subtype'] = '{} yeast'.format(props_new['form_text'])

        elif ingredient_type == 'hops':
            form_code = self.correct_type(props_new['form'])
            form_list = ['pellet', 'plug', 'leaf', 'extract (CO2)', 'extract (isomerized)']
            props_new['form_text'] = form_list[form_code]
            props_new['subtype'] = '{} hops'.format(props_new['form_text'])
            
        elif ingredient_type == 'misc':
            subtype_code = self.correct_type(props_new['type'])
            subtype_list = ['spice', 'fining', 'herb', 'flavor', 'other', 'water-agent']
            props_new['subtype'] = subtype_list[subtype_code]
        else:
            props_new['subtype'] = None

        # process notes
        notes = props_new['notes']
        if notes:
            notes_props = self.process_notes(notes)
            if type(notes_props) in (dict, OrderedDict):
                props_new.update(notes_props)

        # add 'type' tag
        props_new['ingredient_type'] = ingredient_type

        # replace the ingredients
        return props_new

    def process_notes(self, notes):
        # print('notes: {}'.format(notes))

        try:
            props = hjson.loads(notes)
            # print('hjson: {}'.format(props))
        except hjson.scanner.HjsonDecodeError:
            props = notes

        return props

    def flatten_data(self, props, tag_old, tag_new=None):
        props_data = props.pop('data')
        props_dict = self.process_items(props_data, tag_old, tag_new)
        return props_dict

    def strip_key_prefixes(self, prefix, props_from, props_to):
        # print('strip_key_prefixes()-->props_from: {}'.format(props_from))
        if type(props_from) in (dict, OrderedDict):
            for key in props_from:
                # if prefix == 'f_ms_':
                # 	print('strip_key_prefixes()-->key: {}'.format(key))
                # remove key prefix
                if key.startswith(prefix):
                    props_to[key[len(prefix):]] = props_from[key]
                # copy over all other keys
                else:
                    props_to[key] = props_from[key]
        else:
            print('strip_key_prefixes()-->cannot process {}'.format(type(props_from)))

    def strip_key_prefixes_list(self, prefix, props):
        item_num = 0
        while item_num < len(props):
            item_new = {}
            item = props[item_num]
            self.strip_key_prefixes(prefix, item, item_new)
            props[item_num] = item_new
            item_num += 1

    @classmethod
    def get_ingredient_amount(cls, ingredient):
        amount = float(ingredient['amount'])
        units = ingredient['units']
        ingredient_type = ingredient['ingredient_type']

        if ingredient_type == 'grain':
            if units == 'oz':
                amount /= 16
                units = 'lbs'
            
        return amount, units

    def read_xml():
        """
        Reads a BeerXML file and returns a RecipeFolder of recipes
        """
        pass

    def correct_type(self, invar):
        # initialize return variable
        outvar = None

        if type(invar) in (dict, OrderedDict):
            outvar = self.correct_type_dict(invar)
        elif type(invar) is list:
            outvar = self.correct_type_list(invar)

        else:
            try:
                float(invar)
                outvar = float(invar)
                if outvar.is_integer():
                    outvar = int(outvar)
            except (ValueError, TypeError):
                outvar = invar

        return outvar

    def correct_type_dict(self, invar):
        # initialize return variable
        outvar = {}

        for key in invar:
            outvar[key] = self.correct_type(invar[key])

        return outvar

    def correct_type_list(self, invar):
        # initialize return variable
        outvar = []

        index = 0
        while index < len(invar):
            outvar.append(self.correct_type(invar[index]))
            index += 1

        return outvar



import pytest
import os

from beersmith_direct import Recipes

RECIPE_NAME = '2021-11-16_Reston Red Ale'
RECIPE_NAME_2 = '2021-11-16_Reston Red Ale 2'

@pytest.fixture(name='recipes')
def fixture_recipes():
    """Pytest fixture to initialize and return the BeersmithInterface object.
    """
    return Recipes()

def test_read_bsmx_one_recipe(recipes):
    filename = 'bsm-one-recipe.bsmx'
    path = os.path.join(os.getcwd(), 'tests')
    recipe_list = recipes.read_recipes(filename=filename, path=path)

    assert recipe_list
    assert len(recipe_list) == 1
    assert recipe_list[0]['name'] == RECIPE_NAME

def test_read_bsmx_two_recipes(recipes):
    filename = 'bsm-two-recipes.bsmx'
    path = os.path.join(os.getcwd(), 'tests')
    recipe_list = recipes.read_recipes(filename=filename, path=path)

    assert recipe_list
    assert len(recipe_list) == 2
    assert recipe_list[0]['name'] == RECIPE_NAME

def test_read_bsmx_one_folder(recipes):
    filename = 'bsm-one-folder-one-recipe.bsmx'
    path = os.path.join(os.getcwd(), 'tests')
    recipe_list = recipes.read_recipes(filename=filename, path=path)

    assert recipe_list
    assert len(recipe_list) == 1
    assert recipe_list[0]['name'] == RECIPE_NAME

def test_read_bsmx_two_folders(recipes):
    filename = 'bsm-two-folders.bsmx'
    path = os.path.join(os.getcwd(), 'tests')
    recipe_list = recipes.read_recipes(filename=filename, path=path)

    assert recipe_list
    assert len(recipe_list) == 2
    assert recipe_list[0]['name'] == RECIPE_NAME

def test_read_bsmx_nested_folders(recipes):
    filename = 'bsm-nested-folders.bsmx'
    path = os.path.join(os.getcwd(), 'tests')
    recipe_list = recipes.read_recipes(filename=filename, path=path)

    assert recipe_list
    assert len(recipe_list) == 2
    assert recipe_list[0]['name'] == RECIPE_NAME

def test_reset_recipes(recipes):
    recipes.reset()

    assert True

def test_rebuild_one_recipe(recipes):
    filename = 'bsm-one-recipe.bsmx'
    path = os.path.join(os.getcwd(), 'tests')
    recipes.rebuild(filename, path)

    assert recipes.collection.count_documents({}) == 1

def test_rebuild_two_recipes(recipes):
    filename = 'bsm-two-recipes.bsmx'
    path = os.path.join(os.getcwd(), 'tests')
    recipes.rebuild(filename, path)

    assert recipes.collection.count_documents({}) == 2

def test_update_recipes(recipes):
    filename = 'bsm-one-recipe.bsmx'
    path = os.path.join(os.getcwd(), 'tests')

    recipes.reset()
    recipes.update_recipes(filename, path)

    assert recipes.collection.count_documents({}) == 1

@pytest.mark.skip(reason="manual test")
def test_rebuild_complete(recipes):
    """This tests rebuilding the entire database.

    It should not be an automated test.
    """
    recipes.rebuild()

    assert True

def test_verify_config(recipes):
    assert recipes.props.last_updated
    assert recipes.props.last_id

def test_read_rebuild_prop(recipes):
    assert 'rebuild' in recipes.props.props

def test_read_archive_one_action(recipes):
    filename = 'bsm-archive-add.bsmx'
    path = os.path.join(os.getcwd(), 'tests')

    archive_list = recipes.read_archive(filename=filename, path=path)

    assert len(archive_list) == 1

def test_read_archive_two_actions(recipes):
    filename = 'bsm-archive-two-actions.bsmx'
    path = os.path.join(os.getcwd(), 'tests')

    archive_list = recipes.read_archive(filename=filename, path=path)

    assert len(archive_list) == 2

def test_read_archive_add(recipes):
    filename = 'bsm-archive-add.bsmx'
    path = os.path.join(os.getcwd(), 'tests')

    recipes.reset()
    archive_list = recipes.update_recipes_from_archive(filename=filename, basepath=path)

    assert recipes.collection.count_documents({}) == 1

def test_read_archive_edit(recipes):
    filename = 'bsm-archive-edit.bsmx'
    path = os.path.join(os.getcwd(), 'tests')

    recipe_count = recipes.collection.count_documents({})
    recipe = recipes.collection.find_one({'_id': RECIPE_NAME})
    brewer = recipe['brewer']

    recipes.update_recipes_from_archive(filename=filename, basepath=path)

    recipe = recipes.collection.find_one({'_id': RECIPE_NAME})
    assert recipes.collection.count_documents({}) == recipe_count
    assert recipe['brewer'] == 'New Brewer' != brewer

def test_read_archive_delete(recipes):
    filename = 'bsm-archive-delete.bsmx'
    path = os.path.join(os.getcwd(), 'tests')

    recipe_count = recipes.collection.count_documents({})

    recipes.update_recipes_from_archive(filename=filename, basepath=path)

    recipe = recipes.collection.find_one({'_id': RECIPE_NAME})
    assert not recipe
    assert recipes.collection.count_documents({}) == recipe_count - 1

def test_read_archive_paste(recipes):
    filename = 'bsm-archive-paste.bsmx'
    path = os.path.join(os.getcwd(), 'tests')

    recipes.props.rebuild = False
    recipes.props.update()

    recipes.update_recipes_from_archive(filename=filename, basepath=path)

    assert recipes.props.rebuild == True

def test_pull_filter_0(recipes):
    filename = 'bsm-two-recipes.bsmx'
    path = os.path.join(os.getcwd(), 'tests')

    recipes.rebuild(filename, path)
    assert recipes.collection.count_documents({}) == 2

    filename = 'bsm-archive-two-actions.bsmx'
    begin_str = '2021-11-24'
    recipes.pull(filename=filename, path=path, begin_str=begin_str)

    recipe = recipes.collection.find_one({'_id': RECIPE_NAME})
    assert recipe['brewer'] == 'Romano'
    recipe = recipes.collection.find_one({'_id': RECIPE_NAME_2})
    assert recipe['brewer'] == 'Romano'

def test_pull_filter_1(recipes):
    filename = 'bsm-two-recipes.bsmx'
    path = os.path.join(os.getcwd(), 'tests')

    recipes.rebuild(filename, path)
    assert recipes.collection.count_documents({}) == 2

    filename = 'bsm-archive-two-actions.bsmx'
    begin_str = '2021-11-15'
    recipes.pull(filename=filename, path=path, begin_str=begin_str)

    recipe = recipes.collection.find_one({'_id': RECIPE_NAME})
    assert recipe['brewer'] == 'Romano'
    recipe = recipes.collection.find_one({'_id': RECIPE_NAME_2})
    assert recipe['brewer'] == 'New Brewer'

def test_pull_filter_2(recipes):
    filename = 'bsm-two-recipes.bsmx'
    path = os.path.join(os.getcwd(), 'tests')

    recipes.rebuild(filename, path)
    assert recipes.collection.count_documents({}) == 2

    filename = 'bsm-archive-two-actions.bsmx'
    begin_str = '2021-10-31'
    recipes.pull(filename=filename, path=path, begin_str=begin_str)

    recipe = recipes.collection.find_one({'_id': RECIPE_NAME})
    assert recipe['brewer'] == 'New Brewer'
    recipe = recipes.collection.find_one({'_id': RECIPE_NAME_2})
    assert recipe['brewer'] == 'New Brewer'

def test_read_archive_official(recipes):
    filename = 'Archive.bsmx'
    # path = os.path.join(os.getcwd(), 'tests')
    path = None

    archive_list = recipes.read_archive(filename=filename)

    assert len(archive_list) > 0



import os

import pytest

from beersmith_direct.connector import Connector


def test_init_connector():
    connector = Connector(config_name='beersmith')

    assert connector

@pytest.fixture(name='conn')
def fixture_connector():
    """Pytest fixture to initialize and return a Connector.
    """
    connector = Connector(config_name='beersmith')

    return connector

def test_init_connector_beersmith(conn):
    """Tests the Connector's BeerSmith initialization.
    """
    # test initialization
    beersmith_path = os.environ.get('BEERSMITH_DEFAULT_PATH')
    assert conn.bsm.default_path == beersmith_path

def test_init_connector_mongodb(conn):
    """Tests the Connector's MongoDB initialization.
    """
    # test authentication
    mdb = conn.mdb
    assert mdb

    # verify database name
    db_name = os.environ.get('MONGODB_DBNAME')
    assert mdb.name == db_name



"""A set of functions to retrieve and save Beersmith data into MongoDB.
"""

# import main modules here, e.g., 
from beersmith_direct.i_beersmith import BeersmithInterface
from beersmith_direct.recipes import Recipes


__version__ = '1.0.1'



"""Class module that connects to both Beersmith and MongoDB.
"""
from datetime import datetime
import os

from aracnid_config import Config
from aracnid_logger import Logger
from aracnid_utils import timespan as ts
from i_mongodb import MongoDBInterface

from beersmith_direct.i_beersmith import BeersmithInterface

# initialize logging
logger = Logger(__name__).get_logger()

logger.debug('module installed')


class Connector(MongoDBInterface):
    """Provides interfaces to Beersmith and MongoDB to enable data exchange.

    This class can be used on its own or inherited, but it is more useful to
    create subclasses that inherit from the Connector class for specific data
    type. The attributes and instance methods help with this.

    Environment Variables:
        TBD

    Attributes:
        config_name: Name of the configuration object in MongoDB.
        props: Configuration Properties object.
    """

    def __init__(self, config_name=None):
        """Initializes the interfaces and instance attributes.
        """
        self.bsm = BeersmithInterface()
        self.mdbi = MongoDBInterface.__init__(self)

        self.config_name = config_name
        if config_name:
            logger.debug(f'config_name: {self.config_name}')
            self.props = Config(self.config_name, mdb=self.mdbi)

        self.set_start_min()

    def set_start_min(self):
        """Sets an attribute for the minimum start time.

        The minimum start time is determined by the environment.

        Args:
            None
        """
        start_str = os.environ.get('BEERSMITH_START_STR')
        self.start_min = datetime.fromisoformat(start_str).astimezone()

    def timespan(self, **kwargs):
        """Calculates the endpoints of a timespan.

        This instance method supplies the generic timespan function with more
        specific start times based on the last updated order or a preset
        minimum start time.

        Args:
            kwargs: Keyword arguments that specify the timespan.

        Returns:
            The start and end datetime objects that define the timespan.
        """
        begin = kwargs.get('begin')
        begin_str = kwargs.get('begin_str')
        week_str = kwargs.get('week_str')
        if not begin and not begin_str and not week_str:
            kwargs['begin'] = self.datetime_begin()

        return ts(**kwargs)

    def datetime_begin(self):
        """Provides the start time based on the last updated order.

        If no orders have been saved, this instance method defaults to the
        minimum start time preset from the environment.

        Returns:
            Datetime object that represents the start time of the timespan.
        """
        last_updated = self.props.last_updated

        if last_updated:
            return last_updated

        return self.start_min



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['beersmith_direct']

package_data = \
{'': ['*']}

install_requires = \
['aracnid-config',
 'aracnid-logger',
 'aracnid-utils',
 'beautifulsoup4',
 'hjson',
 'i-mongodb',
 'xmltodict']

setup(name='beersmith_direct',
      version='1.0.1',
      description='A set of functions to retrieve and save Beersmith data into MongoDB.',
      author='Jason Romano',
      author_email='aracnid@gmail.com',
      url='https://github.com/lakeannebrewhouse/beersmith-direct',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.9',
     )



"""Recipes classes.
"""
import os.path

from dataclasses import dataclass
from datetime import datetime

from aracnid_logger import Logger
from dateutil.parser import parse
from pymongo.collection import ReturnDocument

from beersmith_direct.connector import Connector

# initialize logging
logger = Logger(__name__).get_logger()


class Recipes(Connector):
    """Contains the code to connect and process recipes from Beersmith.
    """
    def __init__(self) -> None:
        """Initializes the Recipes Connector.

        Establishes connections to Beersmith and MongoDB.
        Sets up access to configuration properties.

        """
        self.collection_name = os.environ.get('BEERSMITH_COLLECTION')
        self.collection_name_raw = f'raw_{self.collection_name}'
        logger.debug(f'collection_name: {self.collection_name}')
        super().__init__(config_name=self.collection_name)

        # initialize MongoDB collection
        self.collection = self.read_collection(self.collection_name)
        self.collection_raw = self.read_collection(self.collection_name_raw)

    def pull(self, filename=None, path=None, save_last=True, **kwargs):
        """Pull updated recipes in MongoDB.

        Args:
            filename: Name of the recipe file.
            path: Location of the recipe file.
            save_last (bool): if set to True (default), details of the last
                object retrieved is saved in the configuration properties
            **kwargs: keyword arguments that specify the timespan to retrieve.
        """
        start, end = self.timespan(collection='beersmith', **kwargs)
        logger.debug(f'timespan: {start}, {end}')

        archive_list = self.update_recipes_from_archive(
            filename=filename, basepath=path, start=start, end=end
        )

        # clear rebuild flag, if set
        self.props.rebuild = False
        self.props.update()

        # logger.debug(f'recipes processed: {updated_count}')

    def rebuild(self, filename=None, path=None, save_last=True):
        """Rebuild recipes in MongoDB.

        Args:
            filename: Name of the recipe file.
            path: Location of the recipe file.
            save_last (bool): if set to True (default), details of the last
                object retrieved is saved in the configuration properties
        """
        # reset the database
        self.reset()

        updated_count = self.update_recipes(filename, path, save_last)

        # clear rebuild flag, if set
        self.props.rebuild = False
        self.props.update()

        # verify count
        recipe_count = self.collection.count_documents({})
        if recipe_count == updated_count:
            logger.debug(f'Recipe collection rebuilt, count: {recipe_count}')
        else:
            logger.warning('Recipe collection rebuilt, discrepancies found')

    def reset(self):
        logger.debug('Resetting recipe collection...')
        self.collection.drop()

        # reset props
        self.props.delete()

    def read_recipes(self, filename=None, path=None, **kwargs):
        """Returns a set of recipes.

        Args:
            filename: Name of the recipe file.
            path: Location of the recipe file.
            **kwargs: keyword arguments that specify the timespan to retrieve.
        """
        start, end = self.timespan(collection='beersmith', **kwargs)
        logger.debug(f'timespan: {start}, {end}')

        recipe_list = self.bsm.read_bsmx(filename=filename, path=path)

        return recipe_list

    def read_archive(self, filename=None, path=None):
        """Returns a set of recipes from the recipe archive.

        Args:
            filename: Name of the recipe file.
            path: Location of the recipe file.
            start: Beginning date to process archive records.
            end: End date to process archive records.
        """
        archive_list = self.bsm.read_bsmx(filename=filename, path=path)

        return archive_list

    def save_raw_recipe(self, recipe):
        pass

    def update_recipes(self, filename=None, path=None, save_last=True):
        """Update recipes in MongoDB.

        Args:
            filename: Name of the recipe file.
            path: Location of the recipe file.
            save_last (bool): if set to True (default), details of the last
                object retrieved is saved in the configuration properties
        """

        # read the recipes from BeerSmith
        recipe_list = self.read_recipes(filename, path)

        # update each recipe
        update_count = 0
        for update_count, recipe in enumerate(recipe_list):
            self.update_recipe(recipe)

            # update config properties
            if save_last:
                self.props.last_updated = datetime.now().astimezone()
                self.props.last_id = recipe['name']
                self.props.update()

        return update_count + 1

    def update_recipe(self, recipe):
        """Save the provided BeerSmith Recipe into MongoDB.

        Args:
            recipe: Beersmith Recipe.

        Returns:
            The MongoDB representation of the BeerSmith Recipe object.
        """
        recipe_id = recipe['name']
        updated_recipe = self.collection.find_one_and_replace(
            filter={'_id': recipe_id},
            replacement=recipe,
            upsert=True,
            return_document=ReturnDocument.AFTER
        )

        return updated_recipe

    def delete_recipe(self, recipe_id=None):
        """Delete the recipe specified by the recipe identifier.

        Args:
            recipe: Beersmith Recipe identifier.

        Returns:
            The MongoDB representation of the BeerSmith Recipe object.
        """
        self.collection.delete_one(
            filter={'_id': recipe_id}
        )

    def update_recipes_from_archive(self, 
        filename='Archive.bsmx', basepath=None, start=None, end=None):
        """Updates a set of recipes from the recipe archive.

        Args:
            filename: Name of the archive file.
            basepath: Location of the archive file.
            start: Beginning date to process archive records.
            end: End date to process archive records.
        """
        rebuild_recipes = False

        # read the archive
        archive_list = self.read_archive(filename, basepath)

        for archive in archive_list:
            action_date = archive['date']

            if start and action_date < start.date().isoformat():
                continue

            if end and action_date > end.date().isoformat():
                continue
            
            action = archive['action']
            filename = archive['file']
            directory = archive['directory']
            archive_date = parse(archive['date'])
            logger.debug(f'[{archive_date}] {archive["name"]}: {action}')

            # set recipe file path
            if directory.startswith('/'):
                directory = directory[1:]
            path = os.path.join(basepath, directory)

            if action in ('Add Recipe', 'Insert/Paste'):
                recipe_list = self.read_recipes(filename, path)
                recipe = recipe_list[0]
                self.update_recipe(recipe)

                logger.debug(f'\tprocessed: {action}')

            elif action in ('Edit', 'Move'):
                recipe_list = self.read_recipes(filename, path)
                recipe = recipe_list[0]

                # confirm the recipe already exists
                recipe_id = recipe['name']
                recipe_check = self.collection.find_one({'_id': recipe_id})
                if recipe_check:
                    self.update_recipe(recipe)
                    logger.debug(f'\tprocessed: {action}')

                # if recipe was renamed, there is no way to find which was the source recipe
                # need to reload the entire database
                else:
                    logger.debug('\tEditing recipe names cause database inconsistencies. The database will be rebuilt.')
                    # config.rebuild = True
                    rebuild_recipes = True  # should be able to rebuild automatically, unless renamed recipe is not unique
                    break

            elif action == 'Delete/Cut':
                recipe_id = archive['name']
                self.delete_recipe(recipe_id)

            elif action == 'Paste':
                # this will create a twin record and there's no way of differentiating it with the original
                # need to reload the entire database
                logger.warning('\tCannot process "Paste" actions. Confirm that no duplicate recipes exist and rebuild the database.')
                rebuild_recipes = True  # need to let user fix the condition first and manually rebuild
                break

        # reload the entire database if necessary
        if rebuild_recipes:
            self.props.rebuild = True
            self.props.update()


@dataclass
class Recipe:
    name: str = 'default'

    def __init__(self) -> None:
        pass
